header {
package com.javadude.xl2;

import java.io.*;
}

// @@parser
//-----------------------------------------------------------------------------
// Define a Parser, calling it XLRecognizer
//-----------------------------------------------------------------------------
class XLRecognizer extends Parser;
options {
  defaultErrorHandler = true;      // Don't generate parser error handlers
}


// Define some methods and variables to use in the generated parser.
{
  // Define a main
  public static void main(String[] args) {
    // Use a try/catch block for parser exceptions
    try {
      // if we have at least one command-line argument
      if (args.length > 0 ) {
        System.err.println("Parsing...");

        // for each directory/file specified on the command line
        for(int i=0; i< args.length;i++)
          doFile(new File(args[i])); // parse it
      }
      else
        System.err.println("Usage: java XLRecogizer <directory name>");

    }
    catch(Exception e) {
      System.err.println("exception: "+e);
      e.printStackTrace(System.err);   // so we can get stack trace
    }
  }


  // This method decides what action to take based on the type of
  //   file we are looking at
  public static void doFile(File f) throws Exception {
    // If this is a directory, walk each file/dir in that directory
    if (f.isDirectory()) {
      String files[] = f.list();
      for(int i=0; i < files.length; i++)
        doFile(new File(f, files[i]));
    }

    // otherwise, if this is a java file, parse it!
    else if ((f.getName().length()>5) &&
             f.getName().substring(f.getName().length()-3).equals(".xl")) {
      System.err.println("-------------------------------------------");
      System.err.println(f.getAbsolutePath());
      parseFile(new FileInputStream(f));
    }
  }

  // Here's where we do the real work...
  public static void parseFile(InputStream s) throws Exception {
    try {
      // Create a scanner that reads from the input stream passed to us
      XLLexer lexer = new XLLexer(s);

      // Create a parser that reads from the scanner
      XLRecognizer parser = new XLRecognizer(lexer);

      // start parsing at the compilationUnit rule
      parser.program();
    }
    catch (Exception e) {
      System.err.println("parser exception: "+e);
      e.printStackTrace();   // so we can get stack trace		
    }
  }
}

// the following tag is used to find the start of the rules section for
//   automated chunk-grabbing when displaying the page
// @@startrules


program
  : "program" IDENT EQUALS 
      subprogramBody 
    DOT
    // end-of-file
  ;


subprogramBody
  : (basicDecl)*
    (procedureDecl)*
    "begin"
        statementList
    "end" IDENT
  ;


basicDecl
  : varDecl
  | constDecl
  | typeDecl
  ;


varDecl
  : "var" identList COLON typeName
    (BECOMES constantValue)?
    SEMI
  ;


constDecl
  : "constant" identList COLON typeName
    BECOMES constantValue SEMI
  ;



identList
  : IDENT (COMMA IDENT)*
  ;


constantValue
  : INTLIT
  | STRING_LITERAL
  | IDENT
  ;


typeDecl
  : "type" IDENT EQUALS
    ( arrayDecl
    | recordDecl
    )
    SEMI
  ;


arrayDecl
  : "array" LBRACKET integerConstant DOTDOT integerConstant RBRACKET
    "of" typeName
  ;

integerConstant
  : INTLIT
  | IDENT // again, a constant...
  ;


recordDecl
  : "record" (identList COLON typeName SEMI)+ "end" "record"
  ;


typeName
  : IDENT
  | "Integer"
  | "Boolean"
  ;


procedureDecl
  : "procedure" IDENT (formalParameters)? EQUALS
        subprogramBody
    SEMI
  ;


formalParameters
  : LPAREN parameterSpec (SEMI parameterSpec)* RPAREN
  ;


parameterSpec
  : ("var")? identList COLON typeName
  ;

statement
  : exitStatement
  | returnStatement
  | ifStatement
  | loopStatement
  | ioStatement
  | (IDENT (LPAREN|SEMI))=> procedureCallStatement
  | assignmentStatement
  ;


statementList
  : statement  statementList
  | (endStatement)=> endStatement statementList
  | // nothing
  ;

assignmentStatement
  : variableReference BECOMES expression SEMI
  ;


exitStatement
  : "exit" "when" expression SEMI
  ;

procedureCallStatement
  : IDENT (actualParameters)? SEMI
  ;

actualParameters
  : LPAREN (expression (COMMA expression)*)? RPAREN
  ;


returnStatement
  : "return" SEMI
  ;


ifStatement
  : "if" ifPart "end" "if" SEMI
  ;

ifPart
  : expression "then"
    statementList
    ( "elsif" ifPart
    | "else" statementList
    )?
  ;

loopStatement
  : ("while" expression)? "loop"
        statementList
    "end" "loop" SEMI
  ;

endStatement
  : "end" SEMI
  ;

variableReference
  : IDENT
    ( LBRACKET expression RBRACKET
    | DOT IDENT
    )*
  ;



ioStatement
  : "put" LPAREN expression RPAREN SEMI
  | "get" LPAREN variableReference RPAREN SEMI
  | "newLine" (LPAREN RPAREN)? SEMI
  | "skipLine" (LPAREN RPAREN)? SEMI
  ;


primitiveElement
  : variableReference
  | constantValue
  | LPAREN expression RPAREN
  ;


booleanNegationExpression
  : ("not")* primitiveElement
  ;

signExpression
  : ((PLUS|MINUS))* booleanNegationExpression
  ;


multiplyingExpression
  : signExpression ((TIMES|DIV|"mod") signExpression)*
  ;

addingExpression
  : multiplyingExpression ((PLUS|MINUS) multiplyingExpression)*
  ;

relationalExpression
  : addingExpression ((EQUALS|NOT_EQUALS|GT|GTE|LT|LTE) addingExpression)*
  ;

expression
  : relationalExpression (("and"|"or") relationalExpression)*
  ;

// @@endparser
// @@endrules

// @@scanner
//----------------------------------------------------------------------------
// The XL scanner
//----------------------------------------------------------------------------
class XLLexer extends Lexer;

options {
  charVocabulary = '\0'..'\377';
  testLiterals=false;    // don't automatically test for literals
  k=2;                   // two characters of lookahead
}

// @@startrules

// Single-line comments
COMMENT
  : "//" (~('\n'|'\r'))*
    { $setType(Token.SKIP); }
  ;


// Literals
protected DIGIT
  : '0'..'9'
  ;

INTLIT 
  : (DIGIT)+
  ;

CHARLIT
  : '\''! . '\''!
  ;

// string literals
STRING_LITERAL
  : '"'!
    ( '"' '"'!
    | ~('"'|'\n'|'\r')
    )*
    ( '"'!
    | // nothing -- write error message
    )
	;

// Whitespace -- ignored
WS
  : ( ' '
    | '\t'
    | '\f'

    // handle newlines
    | ( "\r\n"  // DOS/Windows
      | '\r'    // Macintosh
      | '\n'    // Unix
      )
      // increment the line count in the scanner
      { newline(); }
    )
    { $setType(Token.SKIP); }
  ;


// an identifier.  Note that testLiterals is set to true!  This means
// that after we match the rule, we look in the literals table to see
// if it's a literal or really an identifer
IDENT
  options {testLiterals=true;}
  : ('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'0'..'9')*
  ;
  
// Operators
DOT        : '.'   ;
BECOMES    : ":="  ;
COLON      : ':'   ;
SEMI       : ';'   ;
COMMA      : ','   ;
EQUALS     : '='   ;
LBRACKET   : '['   ;
RBRACKET   : ']'   ;
DOTDOT     : ".."  ;
LPAREN     : '('   ;
RPAREN     : ')'   ;
NOT_EQUALS : "/="  ;
LT         : '<'   ;
LTE        : "<="  ;
GT         : '>'   ;
GTE        : ">="  ;
PLUS       : '+'   ;
MINUS      : '-'   ;
TIMES      : '*'   ;
DIV        : '/'   ;
// @@endrules

// @@endscanner



